//============================================================================//
//
//  Software Name : MEGADOCK
//
//  Class Name : FFTProcess
//
//  Contact address : Tokyo Institute of Technology, AKIYAMA Lab.
//
//============================================================================//

#ifndef FFTProcess_h
#define FFTProcess_h 1

//#define CUFFT

#include <algorithm>
#include <functional>

#include "cpu_time.h"
#include "parallel.h"
#include "parameter.h"
#include "protein.h"
#include "ligand.h"//shimoda130119
#include "receptor.h"//shimoda130709

#ifdef CUFFT
#include "helper_cuda.h"
#include "cufft.h"
#include <thrust/device_vector.h>
#include <thrust/device_ptr.h>
#endif

#include "fftw3.h"

using namespace std;

typedef struct {
    float score;
    int   index[4];
} SortScore;

class FFTProcess
{
private:
    FFTProcess(FFTProcess &c) {}
    const FFTProcess & operator=(const FFTProcess &c);
#ifdef CUFFT
    //host side
    cufftHandle *cufft_plan;
    cufftResult *cufft_result;
    cufftComplex *_inBuf;
    cufftComplex *_outBuf;
    float **_score;
    int   **_pos;

    //device side
    cufftComplex **_CUFFTin_dev;
    cufftComplex **_CUFFTout_dev;
    float **_FFT_rec_r_dev;
    float **_FFT_rec_i_dev;
    float **_score_dev;
    int   **_pos_dev;

    float **radius_core2_dev;
    float **radius_surf2_dev;
    float **_Charge_dev;
    float **xd_dev;
    float **yd_dev;
    float **zd_dev;
    float **grid_coord_dev;
    float **mol_coord_dev;
    float **grid_r_dev;
    float **grid_i_dev;
    float **mol_coord_orig_dev;
    float **mol_coord_orig_center_dev;
    float **ligand_rotation_angle_dev;

    // for sorting (t > 1)
    vector< thrust::device_vector<float> > temp_score_dev_vec;
    vector< thrust::device_vector<int> > temp_index_dev_vec;
    float **temp_score_dev_ptr;
    int **temp_index_dev_ptr;
    float **temp_score_host;
    int **temp_index_host;

#endif /* CUFFT */

    fftwf_plan *plan_fftw_forward;
    fftwf_plan *plan_fftw_inverse;

    fftwf_complex         *_FFTWin;  //shimoda130517
    fftwf_complex         *_FFTWout;

    CPUTime           *_cputime;
    Parallel          *_parallel;
    Parameter         *_parameter;
    Ligand    *_ligand; //shimoda130119
    Receptor   *_receptor; //shimoda130709
    vector<SortScore>     _Top;
    vector< vector<SortScore> >   _Select;
    int               _Num_fft;
    float             *_FFT_rec_r;
    float             *_FFT_rec_i;
    float             **_FFT_lig_r;
    float             **_FFT_lig_i;
    int               *_Current_rot_angle_num;
protected:
    virtual void              alloc_fft();
    virtual void          fft3d(const float &theta,int myid2);
public:
    FFTProcess(CPUTime *pcputime,Parallel *pparallel,Parameter *pparameter, Receptor *preceptor, Ligand *pligand)
        : _cputime(pcputime),_parallel(pparallel),_parameter(pparameter),_receptor(preceptor),_ligand(pligand) { //shimoda130709
        /*
            FFTProcess(CPUTime *pcputime,Parallel *pparallel,Parameter *pparameter)
                : _cputime(pcputime),_parallel(pparallel),_parameter(pparameter) {
        //*/
#ifdef DEBUG
        cout << "Constructing FFTProcess.\n";
#endif
        //cout << "FFTP const "<< _parameter->_Num_sort <<endl; cout.flush();
    }
    virtual ~FFTProcess() {
#ifdef DEBUG
        cout << "Destructing FFTProcess.\n";
#endif
        delete [] _Current_rot_angle_num;
        delete [] _FFT_rec_r;
        delete [] _FFT_rec_i;
        delete [] _FFT_lig_r;
        delete [] _FFT_lig_i;

#ifdef CUFFT
        //host side
        delete [] cufft_plan;
        delete [] cufft_result;
        delete [] _inBuf;
        delete [] _outBuf;
        delete [] _score;
        delete [] _pos;

        //device side
        delete [] _CUFFTin_dev;
        delete [] _CUFFTout_dev;
        delete [] _FFT_rec_r_dev;
        delete [] _FFT_rec_i_dev;
        delete [] _score_dev;
        delete [] _pos_dev;

        delete [] radius_core2_dev;
        delete [] radius_surf2_dev;
        delete [] _Charge_dev;
        delete [] xd_dev;
        delete [] yd_dev;
        delete [] zd_dev;
        delete [] grid_coord_dev;
        delete [] mol_coord_dev;
        delete [] grid_r_dev;
        delete [] grid_i_dev;

        delete [] mol_coord_orig_dev;
        delete [] mol_coord_orig_center_dev;
        delete [] ligand_rotation_angle_dev;

        vector< thrust::device_vector<float> > tmp_sc;
        vector< thrust::device_vector<int> > tmp_in;
        temp_score_dev_vec.swap(tmp_sc);
        temp_index_dev_vec.swap(tmp_in);
        delete [] temp_score_dev_ptr;
        delete [] temp_index_dev_ptr;
        delete [] temp_score_host;
        delete [] temp_index_host;

#endif

        delete [] plan_fftw_forward;
        delete [] plan_fftw_inverse;

        fftwf_free(_FFTWin);
        fftwf_free(_FFTWout);
        fftwf_cleanup(); //Sato san rev. [FFTW's planner saves some other persistent data]

        vector< vector<SortScore> > tmp1;
        vector<SortScore> tmp2;
        _Select.swap(tmp1);
        _Top.swap(tmp2);
    }
    virtual void      alloc_array(const int &num_fft);
    virtual void      receptor_fft(float *grid_r,float *grid_i);
#ifdef CUFFT
    virtual void      cuda_fft(float *grid_r,float *grid_i,float *grid_coord,float *mol_coord,float *theta, int myid2); //shimoda130120
    virtual void      ligand_voxelization_on_dev(float *theta, int myid2);
    virtual void      ligand_info_transfer(float *grid_coord); //shimoda130430
#endif
    virtual void      ligand_fft(float *grid_r,float *grid_i, int myid2);
    virtual void      fft_score(int myid2);
    virtual void      score_sort(int myid2);
    virtual void      fft_memory_free(); //shimoda130517
    virtual void      top_score_clean();
    virtual int       num_fft() {
        return _Num_fft;
    }
    virtual void      num_fft(const int &i) {
        _Num_fft = i;
    }
    virtual void      rotation_index(int i,const int &j) {
        _Current_rot_angle_num[i] = j;
    }
    virtual float     top_score(const int &j) {
        return _Top[j].score;
    }
    virtual int       top_index(const int &j,int k) {
        return _Top[j].index[k];
    }
    virtual void      sort_index(float *fwork,int *iwork);
};

#endif
